#include "debug.h"
#include "project.h"
#include "ui.h"
#include "konst.fs.h"

CMotorDebugger::CMotorDebugger() {
    active = running = false;
}

CMotorDebugger::~CMotorDebugger() {
}

lrkind CMotorDebugger::init() {
    lrkind res = lrok;

    int fd = open("/dev/tty", O_WRONLY, 0);
    if((ioctl(fd, VT_OPENQRY, &vtno) < 0) || (vtno == -1)) res = lrnotty;
    close(fd);

    string tname = "/dev/tty" + i2str(vtno);

    if((res != lrok) || access(tname.c_str(), W_OK)) {
	res = lrnotty;
    } else if(pathfind("gdb", getenv("PATH") ? getenv("PATH") : "").empty()) {
        res = lrgdbshut;
    } else if(!active) {
        rungdb();
	read();
        command("set confirm off", ckdontanalyze);
        command("set print null-stop", ckdontanalyze);
	if(!active) res = lrgdbshut;
    }

    return res;
}

void CMotorDebugger::rungdb() {
    int inpipe[2], outpipe[2];

    if(!pipe(inpipe) && !pipe(outpipe)) {
	active = true;
        gdbpid = fork();

        if(!gdbpid) {
            core.restorestate();
	    dup2(inpipe[1], STDOUT_FILENO);
	    dup2(inpipe[1], STDERR_FILENO);
	    dup2(outpipe[0], STDIN_FILENO);
            
            close(inpipe[0]);
            close(inpipe[1]);
            close(outpipe[0]);
            close(outpipe[1]);

            string term = "/dev/tty" + i2str(vtno);
            string binname = "\"" + proj->absfname(proj->result) + "\"";

            execlp("gdb", "gdb", binname.c_str(), "-silent", "--fullname", "-tty", term.c_str(), 0);
            _exit(0);
        } else {
            fpr = fdopen(inpipe[0], "r");
            fpw = fdopen(outpipe[1], "w");
        }
    }
}

void CMotorDebugger::setbreakpoint(breakpoint &b) {
    if(b.line == -1) {
        command("break main", ckbreakpoint);
    } else {
        string cmd = "break ";
        if(!b.fname.empty()) {
            cmd += (string) justfname(b.fname.c_str()) + ":" + i2str(b.line);
            command(cmd, ckbreakpoint);
        }
    }

    if(line != -1) b.id = line;
}

void CMotorDebugger::setbreakpoints() {
    vector<breakpoint>::iterator ib;
    for(ib = bpoints.begin(); ib != bpoints.end(); ib++) setbreakpoint(*ib);
}

void CMotorDebugger::done() {
    terminated();
    command("kill");
    active = running = false;
    kill(gdbpid, SIGTERM);
}

string CMotorDebugger::read() {
    bool finished = false;
    int len, fd = fileno(fpr), res;
    struct timeval tv;
    fd_set ds;
    string buf;
    char c;

    for(len = 2; !finished && !feof(fpr); len++) {
        FD_ZERO(&ds);
        FD_SET(fd, &ds);
        tv.tv_sec = 3;
        tv.tv_usec = 0;
        
        if(issuchpid(gdbpid)) {
            res = select(fd+1, &ds, 0, 0, &tv);
        } else {
            if(running) {
                terminated();
                lastresult = lrgdbshut;
                buf = "";
            }

            active = false;
            break;
        }
        
        if(FD_ISSET(fd, &ds)) {
            ::read(fd, &c, 1);
            buf += c;
            finished =
              (buf.size() >= 6) &&
              (buf.substr(buf.size()-6) == "(gdb) ");
        }
    }

    if(finished) buf.resize(buf.size()-6);
    return buf;
}

void CMotorDebugger::terminated() {
    running = false;
    command("delete breakpoints", ckdontanalyze);

    vector<breakpoint>::iterator ib;
    vector<watchexpr>::iterator iw;

    while((ib = find(bpoints.begin(), bpoints.end(), false)) != bpoints.end())
        bpoints.erase(ib);

    for(iw = watches.begin(); iw != watches.end(); iw++) iw->val = "";
}

void CMotorDebugger::command(string c, commandkind kind = ckusual) {
    string dout;
    int npos, mpos, id;
    vector<breakpoint>::iterator bp;
    
    if(active) {
	lastresult = lrunknown;
        fprintf(fpw, "%s\n", c.c_str());
        fflush(fpw);

        dout = read();
        
        if(dout.size() && (kind != ckdontanalyze)) {
	    if((dout.find("Program exited") != -1) ||
            (dout.find("The program no longer exists") != -1) ||
            (dout.find("The program is not being run") != -1)) {
                terminated();
            } else if(dout.find("No executable file specified.") != -1) {
                terminated();
                lastresult = lrunabletostart;
            } else if((npos = dout.find("Program received signal ")) != -1) {
                takelocation(dout);
                dout.replace(0, npos, "");
                if((mpos = dout.find("\n")) != -1) dout.resize(mpos);
		resultstr = dout;
		lastresult = lrsignal;
            } else if(dout.find("Breakpoint ") != -1) {
                if((kind != ckbreakpoint) && dout.size() && ((npos = dout.find("\032")) != -1)) {
                    if((mpos = dout.find("Breakpoint ")) != -1)
                    if((bp = find(bpoints.begin(), bpoints.end(),
                    id = atol(dout.substr(mpos+11).c_str()))) !=
                    bpoints.end()) {
                        takelocation(dout);
                    
                        if(bp->permanent) {
                            bp->fname = fname;
                            bp->line = line;
                        } else {
                            bpoints.erase(bp);
                            command("delete " + i2str(id), ckusual);
                        }
                    }
                } else {
                    if((npos = dout.find("Breakpoint ")) != -1) {
                        line = atol(dout.substr(npos+11).c_str());
                    } else {
                        line = -1;
                    }
                }
            } else switch(kind) {
                case cknext:
                    if(dout.find("\032\032") == -1) {
                        command(c, kind);
                    } else {
                        takelocation(dout);
                    }
                    break;
                case ckwatch:
                    if((npos = dout.find("$")) != -1) {
                        dout = dout.substr(npos);
                        if((npos = dout.find(" = ")) != -1) {
                            dout = dout.substr(npos+3);
                            if((mpos = dout.find("\n")) != -1) dout.resize(mpos);
                            watchplace = dout;
                        }
                    }
                    break;
                case ckstack: parsestack(dout); break;
                case ckcore: parsecore(dout); break;
                default: lastresult = lrunknown; break;
            }
        }
    }
}

void CMotorDebugger::takelocation(string p) {
    int npos, mpos;
    strstream liner;
    string s;

    liner << p;

    while(!liner.eof()) {
        getline(liner, s);

        if((npos = s.find(" at ")) != -1) {
            if(s.substr(npos).find(":") != -1) {
                fname = s.substr(npos+4);
            }
        } else if((npos = s.find(":beg:")) != -1) {
            fname = s;
            while(!fname.empty() && (fname[0] != '/')) fname.replace(0, 1, "");
        }
    }

    if((mpos = fname.find(":")) != -1) {
        line = strtoul(fname.substr(mpos+1).c_str(), 0, 0);
        fname.resize(mpos);
    }

    if(proj->watchactive) updatewatch();
}

void CMotorDebugger::parsestack(string stackinfo) {
    strstream liner;
    string s;
    vector<progstackitem>::iterator ip = stack.end();

    liner << stackinfo;
    
    while(!liner.eof()) {
        getline(liner, s);

        if(s[0] == '#') {
            stack.push_back(progstackitem(s));
            ip = stack.end()-1;
        } else if(ip != stack.end()) {
            ip->nextline(s);
        }
    }
}

void CMotorDebugger::parsecore(string info) {
    strstream liner;
    string s;
    bool found = false;

    resultstr = "";
    lastresult = lrcoreok;
    liner << info;

    while(!liner.eof()) {
        getline(liner, s);
        
        if(!s.empty())
        if(s[0] == '#') {
            found = true;
            if(s.find(" ?? ") != -1) lastresult = lrwrongcore;
        }

        resultstr += s + "\n";
    }

    if(!found) lastresult = lrwrongcore;
}

void CMotorDebugger::updatewatch() {
    vector<watchexpr>::iterator iw;

    for(iw = watches.begin(); iw != watches.end(); iw++)
        iw->val = getvar(iw->name);
}

void CMotorDebugger::run() {
    setvt();

    if(!running) {
	running = true;
        setbreakpoints();
	command("run " + proj->args, ckrun);
    } else {
	command("continue", ckcontinue);
    }

    backvt();
}

void CMotorDebugger::loadcore(string corefname) {
    if(!active) {
        init();
    }
    
    command("core " + corefname, ckcore);
    running = true;
}

void CMotorDebugger::readstack(vector<progstackitem> &astack) {
    stack.clear();
    command("info stack", ckstack);
    astack = stack;
}

void CMotorDebugger::setvt() {
    sscanf(getcurtty(), "/dev/tty%d", &thisvt);
    ioctl(vtfd = open("/dev/tty", O_WRONLY, 0), VT_ACTIVATE, vtno);
    ioctl(vtfd, VT_WAITACTIVE, vtno);
}

void CMotorDebugger::backvt() {
    ioctl(vtfd, VT_ACTIVATE, thisvt);
    ioctl(vtfd, VT_WAITACTIVE, thisvt);
    close(vtfd);
}

void CMotorDebugger::proceed(bool into) {
    if(active) {
        setvt();
        command(into ? "s" : "n", cknext);
	backvt();
    }
}

void CMotorDebugger::addbreakpoint(const breakpoint &bp) {
    bpoints.push_back(bp);
    if(running) setbreakpoint(*(bpoints.end()-1));
}

void CMotorDebugger::delbreakpoint(const breakpoint bp) {
    vector<breakpoint>::iterator ib;

    while((ib = find(bpoints.begin(), bpoints.end(), bp)) != bpoints.end()) {
        if(running) command("delete " + i2str(ib->id), ckusual);
        bpoints.erase(ib);
    }
}

void CMotorDebugger::setvar(string name, string newval) {
    command((string) "set var " + name + "=" + newval, ckdontanalyze);
}

string CMotorDebugger::getvar(string varname) {
    watchplace = "";
    command((string) "print " + varname, ckwatch);
    return watchplace;
}

lrkind CMotorDebugger::getlastresult() {
    return lastresult;
}

string CMotorDebugger::getresultstr() {
    return resultstr;
}

string CMotorDebugger::getfname() {
    return fname;
}

int CMotorDebugger::getfline() {
    return line;
}

void CMotorDebugger::clear() {
    bpoints.clear();
    watches.clear();
}

