#include "ui.h"
#include "debug.h"
#include "debugp.h"
#include "edit.h"
#include "project.h"
#include "compile.h"
#include "conf.h"

bool watchstay = false;

void CMotorCore::checkdebugresult() {
    static int prevline = -1, fn;
    vector<progstackitem> stack;
    int i;

    if(deb.running) {
        switch(deb.getlastresult()) {
            case lrsignal: bottom(deb.getresultstr()); break;
            case lrcoreok: bottom(_("Core loaded ok")); break;
            default:
	        bottom(_("gdb: program stopped"));
                break;
	}

        for(i = (deb.getfname().size() ? -1 : 0); ; i++) {
            if(i >= 0) {
                if(!i) deb.readstack(stack);
                if(stack.empty()) break;

                curdebfname = stack.begin()->getfname();
                curdebline = stack.begin()->getline();
                stack.erase(stack.begin());
            } else {
                curdebfname = deb.getfname();
                curdebline = deb.getfline();
            }
        
            if(ed.openfile(curdebfname, curdebline)) {
    	        ed.unlight(fn, prevline);
            
                if(deb.running)
                    ed.enlight(fn = ed.getcurrentfnum(),
                    prevline = curdebline, cp_debug_current);

                break;
            }
        }
    } else {
        switch(deb.getlastresult()) {
            case lrunabletostart: bottom(_("Unable to start the binary")); break;
            case lrgdbshut: bottom(_("Unexpected gdb shutdown")); break;
            default: bottom(_("Terminated ok")); break;
        }
        
        ed.unlight(fn, prevline);
        prevline = -1;
    }

    updatewatches();
    ed.redraw();
}

bool CMotorCore::checkbuildneed() {
    bool r = true;
    static bool askrebuild = false;

    ed.saveall();
    
    if(proj->modified() || access(proj->result.c_str(), F_OK)) {
        if(deb.running) {
            if(!askrebuild) {
                if(ask("YN", _("source was modified. rebuild?")) == ANSWER_NO) {
                    askrebuild = true;
                } else {
                    r = build();
                    if(r && cmp.nfiles) deb.done();
                }
            }
        } else {
            r = build();
        }
    }
    return r;
}

int CMotorCore::execbin() {
    int cpid, status, i;
    string binname = proj->outdir + "/" + proj->result;

    endwin();

    if(!(cpid = fork())) {
	for(i = 0; i < LINES; i++) printf("\n\n");
	fflush(stdout);
        execlp(binname.c_str(), binname.c_str(), proj->args.c_str(), 0);
        _exit(1);
    } else {
        string s;
	wait(&status);
        cout << _("Press Enter to continue..") << flush;
        getline(cin, s);
	bottom(_("Terminated ") + i2str(status));
        redraw();
        return status;
    }
}

void CMotorCore::debugproceed(bool into) {
    if(checkbuildneed()) {
        if(!deb.running) {
            deb.addbreakpoint(breakpoint("", -1, false));
            run();
        } else {
            bottom(_("Continuing running ") + proj->result);
            deb.proceed(into);
            checkdebugresult();
        }
    }
}

void CMotorCore::debugtocursor() {
    int i;

    if(core.checkbuildneed()) {
        ed.getpos(0, &i);
        deb.addbreakpoint(breakpoint(ed.getcurrentfname(), i+1, false));
        core.run();
    }
}

void CMotorCore::debugcore() {
    dialogbox db;
    string corefname = proj->rootdir + "/core";
    
    corefname = input(corefname, _("core dump file name: "));
    
    if(!corefname.empty()) {
        bottom(_("Loading core dump.."));
        deb.loadcore(corefname);
        bottom();

        db.setwindow(new textwindow(0, 0, DIALOG_WIDTH, DIALOG_HEIGHT,
            conf.getcolor(cp_menufr), TW_CENTERED,
            conf.getcolor(cp_menuhl),
            deb.getlastresult() != lrwrongcore ?
                _(" Core dump loaded ok ") :
                _(" Error loading core dump ")));

        db.setbar(new horizontalbar(conf.getcolor(cp_menu),
            conf.getcolor(cp_menusel), _("Ok"), 0));

        db.setbrowser(new textbrowser(conf.getcolor(cp_menu)));
        db.getbrowser()->setbuf(deb.getresultstr().c_str());

        db.open();
        db.close();

        if(deb.getlastresult() == lrwrongcore) {
            deb.done();
        } else {
            checkdebugresult();
        }
    }
}

bool CMotorCore::nodebask(string msg) {
    bool ret = true;
    static bool nodebasked = false;

    if(!nodebasked)
    if(!deb.bpoints.empty()) {
        if(ask("YN", msg) == ANSWER_NO) {
            bottom();
	    ret = false;
        } else {
            nodebasked = true;
        }
    }

    return ret;
}

void CMotorCore::run() {
    lrkind r;

    if(checkbuildneed()) {
	bottom((deb.running ? _("Continuing running ") : _("Running ")) + proj->result);
	if(!deb.active) r = deb.init();

	switch(r) {
	    case lrok:
		deb.run();
    		checkdebugresult();
		break;
	    case lrnotty:
		if(nodebask(_("No free tty to start a debug session. Run without debugging?")))
		    execbin();
		break;
	    case lrgdbshut:
		if(nodebask(_("Can't find gdb. Run the program without debugging?")))
		    execbin();
		break;
	}
    }
}

void CMotorCore::setargs() {
    proj->args = input(proj->args, _("program arguments: "));
}

void CMotorCore::reset() {
    if(deb.active) {
        deb.done();
        checkdebugresult();
        bottom(_("Program reset"));
    }
}

void CMotorCore::showstackitem(const progstackitem &ps) {
    dialogbox db;
    string text, args;

    db.setwindow(new textwindow(0, 0, DIALOG_WIDTH, DIALOG_HEIGHT,
        conf.getcolor(cp_menufr), TW_CENTERED,
        conf.getcolor(cp_menuhl), _(" Call stack "),
        ps.getfunction().c_str()));

    db.setbar(new horizontalbar(conf.getcolor(cp_menu),
        conf.getcolor(cp_menusel), _("Ok"), 0));

    text += (string) _("Function") + "\n";
    text += "  " +
        ps.getfunction() + "() in " +
        ps.getfname() + ":" +
        i2str(ps.getline()) + "\n\n";

    text += (string) _("Arguments") + "\n";
    args += trailcut(leadcut(ps.getargs(), "("), ")");

    while(!args.empty()) {
        text += "  " + leadcut(getwordquote(args, "\"", ",")) + "\n";
    }

    db.setbrowser(new textbrowser(conf.getcolor(cp_menu)));
    db.getbrowser()->setbuf(text.c_str());

    db.open();
    db.close();
}

void CMotorCore::callstack() {
    bool fin;
    int i, b;
    dialogbox db;
    static int citem = 0, savecount;

    vector<progstackitem> stack;
    vector<progstackitem>::iterator is;

    if(!deb.active) {
        bottom(_("No debug session running"));
        return;
    }

    db.setwindow(new textwindow(0, 0, DIALOG_WIDTH, DIALOG_HEIGHT,
        conf.getcolor(cp_menufr), TW_CENTERED,
        conf.getcolor(cp_menuhl), _(" Call stack ")));

    db.setmenu(new verticalmenu(conf.getcolor(cp_menu),
        conf.getcolor(cp_menusel)));
    
    db.setbar(new horizontalbar(conf.getcolor(cp_menu),
        conf.getcolor(cp_menusel), _("Info"), _("Go to"), 0));

    deb.readstack(stack);
    db.getbar()->item = 1;

    for(is = stack.begin(); is != stack.end(); is++)
        db.getmenu()->additem((string) " " + is->getfunction() + is->getargs());

    if(savecount != stack.size()) {
        savecount = stack.size();
        citem = 0;
    } else {
        db.getmenu()->setpos(citem);
    }

    for(fin = false; !fin; )
    if(!(fin = !db.open(i, b))) {
        is = stack.begin()+i-1;
        
        switch(b) {
            case 0: showstackitem(*is); break;
            case 1: fin = ed.openfile(is->getfname(), is->getline()); break;
        }
    }

    db.close();
    ed.redraw();
}

void CMotorCore::openbreakpoints() {
    int i, n, b, nbpoints;
    bool finished;
    dialogbox db;
    string fname = justfname(ed.getcurrentfname().c_str());
    vector<breakpoint>::iterator ib;

    db.setwindow(new textwindow(0, 0, DIALOG_WIDTH, DIALOG_HEIGHT,
        conf.getcolor(cp_menufr), TW_CENTERED,
        conf.getcolor(cp_menuhl), _(" Breakpoints ")));

    db.settree(new treeview(conf.getcolor(cp_menu), conf.getcolor(cp_menusel),
        conf.getcolor(cp_menu), conf.getcolor(cp_menu)));

    db.setbar(new horizontalbar(conf.getcolor(cp_menu),
        conf.getcolor(cp_menusel), _("Add"), _("Remove"), _("Go to"), 0));

    db.getbar()->item = 2;
    db.addkey(KEY_IC, 0);
    db.addkey(KEY_DC, 1);

    for(finished = false; !finished; ) {
        db.gettree()->clear();
        nbpoints = db.gettree()->addnode(0, conf.getcolor(cp_menuhl), 0, _(" Breakpoints "));
        
        for(ib = deb.bpoints.begin(); ib != deb.bpoints.end(); ib++)
        if(ib->permanent)
            db.gettree()->addleaf(nbpoints, 0, 0, " " +
            proj->relativefname(ib->fname) + " (" +
            i2str(ib->line) + ") ");

        if(db.open(i, b)) {
            if(!b) {
                bottom(_("You can only add breakpoints from the editor window"));
            } else if(i > 1) {
                ib = deb.bpoints.begin()+i-2;
                switch(b) {
                    case 1: deb.delbreakpoint(*ib); break;
                    case 2: finished = ed.openfile(ib->fname, ib->line); break;
                }
            }
        } else {
	    finished = true;
        }
    }

    db.close();
    ed.markbreakpoints(); // just to remove deleted ones
    redraw();
}

void CMotorCore::setbreakpoint() {
    breakpoint bp(ed.getcurrentfname(), ed.getpos()+1);

    if(find(deb.bpoints.begin(), deb.bpoints.end(), bp) != deb.bpoints.end()) {
        deb.delbreakpoint(bp);
    } else {
        deb.addbreakpoint(bp);
    }

    ed.markbreakpoints();
    ed.redraw();
}

int CMotorCore::watchkeys(verticalmenu *m, int k) {
    switch(k) {
        case CTRL('d'):
            watchstay = false;
        case ALT('w'):
            return 0;
            
        case KEY_IC:
        case CTRL('w'):
            core.addwatch(ed.atcursor());
            break;

        case KEY_DC:
            if(!deb.watches.empty()) {
                deb.watches.erase(deb.watches.begin()+m->getpos());
                m->remove(m->getpos());
            }
            break;
    }

    m->redraw();
    return -1;
}

void CMotorCore::openwatch(bool activate) {
    int i, wlines = (int) (LINES*0.3), watchcur;
    vector<watchexpr>::iterator iw;

    if(!wwatch || !activate) {
        wwatch = new textwindow(0, LINES-wlines, COLS-1, LINES-2, conf.getcolor(cp_menufr));
        wwatch->set_title(conf.getcolor(cp_menuhl), _(" Watches "));
        wwatch->open();
        proj->watchactive = true;
        ed.resize(elwatch);
    }

    watchstay = true;

    if(activate) {
        watchmenu = new verticalmenu(1, LINES-wlines+1, COLS-1, LINES-2,
        conf.getcolor(cp_menu), conf.getcolor(cp_menusel));

        watchmenu->otherkeys = &watchkeys;
        watchmenu->clearonfocuslost = true;

        bottom(_("Watches: ~Ins~ add, ~Del~ remove, ~^D~ close, ~A-W~ return to editor"));

        while(1) {
            watchmenu->clear();

            for(iw = deb.watches.begin(); iw != deb.watches.end(); iw++)
                watchmenu->additem(iw->name + ": " + (iw->val.empty() ? _("(not avail)") : iw->val));

            if(!(i = watchmenu->open())) break;
            
            if(deb.active && i) {
                iw = deb.watches.begin()+i-1;
                string ans = core.input(iw->val, _("new value: "));
                if(!ans.empty()) deb.setvar(iw->name, ans);
            }
        }

        watchmenu->getpos(watchstart, watchcur);
        watchmenu->close();
        
        delete watchmenu;
        watchmenu = 0;
    }

    if(!watchstay && activate) {
        ed.resize(elnormal);
        delete wwatch;
        wwatch = 0;
        proj->watchactive = false;
    } else {
        updatewatches();
    }

    bottom();
}

void CMotorCore::addwatch(string wstr) {
    static string ans;
    vector<watchexpr>::iterator iw;

    if(!wstr.empty()) ans = wstr;
    ans = input(ans, _("add watch: "));
    
    if(!ans.empty()) {
        deb.watches.push_back(watchexpr(ans));
        deb.updatewatch();

        if(wwatch) {
            if(watchmenu) {
                iw = deb.watches.end()-1;
                watchmenu->additem(iw->name + ": " +
                    (iw->val.empty() ? _("(not avail)") : iw->val));
            } else {
                updatewatches();
            }
        } else {
            openwatch(false);
        }
    }
}

void CMotorCore::updatewatches() {
    int i, ws = watchstart < 0 ? 0 : watchstart;
    
    if(wwatch) {
        ed.resize(elwatch);
        attrset(conf.getcolor(cp_input));

        for(i = ws; (i < deb.watches.size()) && (i-watchstart < 4); i++)
            wwatch->writef(1, 1+i, "%s: %-*s", deb.watches[i].name.c_str(),
                COLS, deb.watches[i].val.empty() ? _("(not avail)") :
                deb.watches[i].val.c_str());
    }
}

void CMotorCore::evaluate(string expr) {
    int i, b;
    static string exps;
    string val, ans;
    dialogbox db;

    db.setwindow(new textwindow(0, 0, DIALOG_WIDTH, (int) (DIALOG_HEIGHT*0.45),
        conf.getcolor(cp_menufr), TW_CENTERED,
        conf.getcolor(cp_menuhl), _(" Evaluate/Modify ")));

    db.settree(new treeview(conf.getcolor(cp_menu), conf.getcolor(cp_menusel),
        conf.getcolor(cp_menu), conf.getcolor(cp_menu)));

    db.setbar(new horizontalbar(conf.getcolor(cp_menu),
        conf.getcolor(cp_menusel), _("Change"), 0));

    db.gettree()->menu.setpos(1);

    bool finished = false, first = true;

    while(!finished) {
        db.gettree()->clear();
        if(!exps.empty()) val = deb.getvar(exps);

        i = db.gettree()->addnode(0, conf.getcolor(cp_menuhl), 0, _(" Expression "));
        db.gettree()->addleaf(i, 0, 0, " " + exps + " ");

        i = db.gettree()->addnode(0, conf.getcolor(cp_menuhl), 0, _(" Value "));
        db.gettree()->addleaf(i, 0, 0, " " + (val.empty() ? _("not available") : val) + " ");

        if(!expr.empty()) {
            if(!expr.empty()) exps = expr;
            if(finished = !first) continue;
            first = false;
            expr = "";
            i = 2;
        } else {
            if(!db.open(i, b)) break;
        }

        switch(i) {
            case 2:
                ans = core.input(exps, _("expression: "));
                if(!ans.empty()) exps = ans; else finished = true;
                break;
            case 4:
                if(!val.empty()) {
                    ans = core.input(val, _("new value: "));
                    if(!ans.empty()) deb.setvar(exps, ans);
                }
                break;
        }
    }

    db.close();
}

